#include <array>
#include <cstdint>
#include <random>
#include <vector>

// Generate an example mesh describing a 2D truss.
// In this example, the nodes are distributed on a cartesian grid, the topology
// of the mesh is as follows:
/*
 * o-o-o-o-o-o  ^
 * |\|\|\|\|\|  |
 * o-o-o-o-o-o  |  n_nodes_y layers
 * |\|\|\|\|\|  |
 * o-o-o-o-o-o  |
 *
 * ----------->
 * n_nodes_x layers
 *
 */

template < typename Node >
struct Mesh
{
    std::vector< std::array< Node, 2 > > topology;
    std::vector< double >                x, y;
};

template < typename Node = std::uint64_t >
auto generateMesh(size_t n_nodes_x, size_t n_nodes_y, unsigned seed = std::random_device{}()) -> Mesh< Node >
{
    const auto n_nodes = n_nodes_x * n_nodes_y;

    // Topology
    std::vector< std::array< Node, 2 > > topology;
    topology.reserve((n_nodes_x - 1) * (n_nodes_y - 1) * 3 + (n_nodes_x - 1) + (n_nodes_y - 1));
    for (Node j = 0; j < n_nodes_y - 1; ++j)
    {
        for (Node i = 0; i < n_nodes_x - 1; ++i)
        {
            const Node base = n_nodes_x * j + i;
            topology.push_back(std::array< Node, 2 >{base, base + 1});
            topology.push_back(std::array< Node, 2 >{base, base + n_nodes_x});
            topology.push_back(std::array< Node, 2 >{base, base + n_nodes_x + 1});
        }
        topology.push_back(std::array< Node, 2 >{n_nodes_x * (j + 1) - 1, n_nodes_x * (j + 2) - 1});
    }
    for (Node i = 0; i < n_nodes_x - 1; ++i)
        topology.push_back(std::array< Node, 2 >{n_nodes_x * (n_nodes_y - 1) + i, n_nodes_x * (n_nodes_y - 1) + i + 1});

    // Node coords
    std::vector< double > x_unshuffled(n_nodes);
    std::vector< double > y_unshuffled(n_nodes);
    Node                  coord_ind = 0;
    for (Node j = 0; j < n_nodes_y; ++j)
        for (Node i = 0; i < n_nodes_x; ++i)
        {
            x_unshuffled[coord_ind] = i;
            y_unshuffled[coord_ind] = j;
            ++coord_ind;
        }

    // Shuffle
    std::mt19937        prng{seed};
    std::vector< Node > permutation(n_nodes);
    std::iota(permutation.begin(), permutation.end(), 0u);
    std::shuffle(permutation.begin(), permutation.end(), prng);
    std::vector< double > x(n_nodes), y(n_nodes);
    for (size_t i = 0; auto p : permutation)
    {
        x[p] = x_unshuffled[i];
        y[p] = y_unshuffled[i];
        ++i;
    }
    for (auto& [n1, n2] : topology)
    {
        n1 = permutation[n1];
        n2 = permutation[n2];
    }
    std::shuffle(topology.begin(), topology.end(), prng);
    return {std::move(topology), std::move(x), std::move(y)};
}
